/*
   File: dat_parser.c
   Parses dat files

   Copyright 2007 Radboud University of Nijmegen
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Library General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   CVS ID: "$Id: dat_parser.c,v 1.12 2007/11/21 15:52:40 marcs Exp $"
*/

/* system includes */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

/* libabase includes */
#include <abase_repr.h>
#include <abase_error.h>
#include <abase_memalloc.h>
#include <abase_fileutil.h>

/* liblexicon includes */
#include <lxcn_input.h>
#include <lxcn_vocabulary.h>

/* local includes */
#include "options.h"
#include "globals.h"
#include "dyn_array.h"
#include "lif_parser.h"
#include "dat_parser.h"
#include "affix_values.h"
#include "nonterminals.h"
#include "entries.h"

/*
   A dat line has the form
   "WORDFORM" nonterminal name [(PARAMS)] [NUMBER]
*/
/*
   Character reading administration
*/
#define MAX_DAT_LINE_LEN 4095
static char dat_line_buffer[MAX_DAT_LINE_LEN + 1];
static char *dat_fname;
static char *line_ptr;
static int has_errors;
static int linenr;
static FILE *dat;

static void dat_error (char *format,...)
{ va_list argp;
  abs_printf ("File '%s', line %d: ", dat_fname, linenr);
  va_start (argp, format);
  abs_vprintf (format, argp);
  va_end (argp);
  abs_printf ("\n");
  has_errors = 1;
}

static void dat_warning (char *format,...)
{ va_list argp;
  if (!verbose) return;
  abs_printf ("    File '%s', line %d: ", dat_fname, linenr);
  va_start (argp, format);
  abs_vprintf (format, argp);
  va_end (argp);
  abs_printf ("\n");
}

static int ch_is_white_space (char ch)
{ return ((ch == ' ') || (ch == '\n') || (ch == '\f') || (ch == '\r') || (ch == '\t'));
}

/* Read line and eat all trailing white space */
static void read_line ()
{ char *line_end;
  int len;
  line_ptr = fgets (dat_line_buffer, MAX_DAT_LINE_LEN, dat);
  linenr++;
  if (line_ptr == NULL) return;
  len = (int) strlen (line_ptr);
  line_end = line_ptr + len - 1;
  while ((len != 0) && ch_is_white_space (*line_end))
    { line_end--; len--; }
  *++line_end = '\0';
}

/* Opening and closing the dat file */
static void try_open_lexicon_file (char *lexicon_name)
{ dat_fname = abs_new_fmtd_string ("try_open_lexicon_file", "%s.%s", lexicon_name, DAT_SUFFIX);
  dat = abs_fopen (dat_fname, "r");

  /* Prepare line buffer */
  has_errors = 0;
  linenr = 0;
  read_line ();
}

static void close_lexicon_file ()
{ if (has_errors)
    abs_abort ("close_lexicon_file", "Lexicon file '%s' contains errors", dat_fname);
  fclose (dat);
  abs_free (dat_fname, "close_lexicon_file");
}

static void may_skip_white_space ()
{ while (ch_is_white_space (*line_ptr)) line_ptr++;
}

static int is_eof ()
{ return (line_ptr == NULL);
}

static int is_eoln ()
{ return (*line_ptr == '\0');
}

static void should_be_eoln ()
{ if (!is_eoln ())
    dat_error ("End of line expected");
  read_line ();
}

static void skip_eoln ()
{ while (!is_eof () && !is_eoln ()) line_ptr++;
}

static int ahead_letter ()
{ if (('a' <= (*line_ptr)) && (*line_ptr <= 'z')) return (1);
  if (('A' <= (*line_ptr)) && (*line_ptr <= 'Z')) return (1);
  return (0);
}

static int ahead_name_char ()
{ if (('a' <= (*line_ptr)) && (*line_ptr <= 'z')) return (1);
  if (('A' <= (*line_ptr)) && (*line_ptr <= 'Z')) return (1);
  if (('0' <= (*line_ptr)) && (*line_ptr <= '9')) return (1);
  if ((((int) (*line_ptr)) & 0xff) >= 128) return (1);	/* Accept all extended ASCII */
  switch (*line_ptr)
    { case '$':
      case '+':
      case '-':
      case '?':
      case '@':
      case '^':
      case '~':
      case '_': return (1);
      default: break;
    };
  return (0);
}

static int is_char (char ch)
{ if (ch == *line_ptr)
    { line_ptr++;
      may_skip_white_space ();
      return (1);
    };
  return (0);
}

static void should_be_char (char ch)
{ if (is_char (ch)) return;
  dat_error ("Special character '%c'", *line_ptr);
  skip_eoln ();
}

/*
   Basic LL(1) parsing of the dat file
*/
static int is_comment ()
{ if (*line_ptr == '#')
    { read_line ();
      return (1);
    };
  return (0);
}

/*
   Reading of numbers:
   For parameters we must be able to distinguish between a signed number
   and a - as the first character of an affix name. Frequencies should
   always be positive
*/
static int ahead_signed_number ()
{ char *ptr;
  if (*line_ptr != '-') return (0);
  for (ptr = line_ptr + 1; ch_is_white_space (*ptr); ptr++) ;
  return (('0' <= *ptr) && (*ptr <= '9'));
}

static int is_digit (int *ret_val)
{ if (('0' <= (*line_ptr)) && (*line_ptr <= '9'))
    { *ret_val = (int) (((*line_ptr) - '0'));
      line_ptr++;
      return (1);
    };
  return (0);
}

static int is_number (int *ret_nr)
{ int value, digit;
  if (!is_digit (&value)) return (0);
  while (is_digit (&digit)) value = value * 10 + digit;
  may_skip_white_space ();
  *ret_nr = value;
  return (1);
}

static void should_be_number (int *ret_nr)
{ if (is_number (ret_nr)) return;
  dat_error ("Number expected");
  skip_eoln ();
};

static int is_signed_number (int *ret_nr)
{ int value;
  if (ahead_signed_number ())
    { should_be_char ('-');
      should_be_number (&value);
      *ret_nr = -value;
      return (1);
    };
  return (is_number (ret_nr));
}

static void may_be_signed_number (int *ret_nr)
{ if (is_signed_number (ret_nr)) return;
  *ret_nr = 1;
}

static int is_number_parameter (int *index)
{ int value;
  if (!is_signed_number (&value)) return (0);
  *index = register_new_int_affix (value);
  return (1);
}

/*
   Recognizing strings
*/
static int is_string (char *dbuf)
{ char *dptr = dbuf;
  int done = 0;
  if (*line_ptr != '"') return (0);
  line_ptr++;
  while (!done)
    switch (*line_ptr)
      { case '\0': dat_error ("Unterminated string"); done = 1; break;
	case '"':  line_ptr++; done = 1; break;
	case '\\':
	  { line_ptr++;
	    if (is_eoln ()) break;
	    *dptr++ = '\\';
	    *dptr++ = *line_ptr++;
	  }; break;
	default:
	  *dptr++ = *line_ptr++;
      };
  *dptr = '\0';
  may_skip_white_space ();
  return (1);
}

/*
   Recognize strings as wordforms and handle hyphen & layout conventions
   All special marks must be prefixed with a backslash
   Hence: "\-s" denotes a suffix "s"
	  
   Convention: "\-" at start of wordform -> Suffix
	       "\-" at end of wordform -> Prefix
	       "\-" in middle of wordform -> Soft Hyphen
	       "\!" at start of wordform -> Literal match
	       "\n" -> a newline
	       "\t" -> a tab
	       "\r" -> a return
	       "\\" -> a backslash
	       "\"" -> a double quote
	       "\uHHHH"     -> UTF8 encoding of unicode character HHHH
	       "\UHHHHHHHH" -> UTF8 encoding of unicode character HHHHHHHH

   Special markers are orthogonal:
   "\-\!oxy\-" denotes a literal infix "oxy"

   Note: all leading and trailing white space in word forms is removed

   UTF8 encoding:           | 1st byte | 2nd byte | 3rd byte | 4th byte |
   U00000000 - U0000007F -> | 0xxxxxxx |          |          |          |
   U00000080 - U000007FF -> | 110yyyyy | 10xxxxxx |          |          |
   U00000800 - U0000FFFF -> | 1110zzzz | 10yyyyyy | 10xxxxxx |          |
   U00010000 - U001FFFFF -> | 11110uuu | 10uuzzzz | 10yyyyyy | 10zzzzzz |
*/
static int is_hex_digit (char ch, int *val)
{ if (('0' <= ch) && (ch <= '9'))
    { *val = (int) (ch - '0');
      return (1);
    };
  if (('A' <= ch) && (ch <= 'F'))
    { *val = (int) (ch - 'A') + 10;
      return (1);
    };
  if (('a' <= ch) && (ch <= 'f'))
    { *val = (int) (ch - 'a') + 10;
      return (1);
    };
  return (0);
}

static void add_utf8_character (char **sptr, char **dptr, int nr)
{ char *ptr = *sptr;
  char *tptr = *dptr;
  int value = 0;
  int ix, dig, hdr, nr_bytes;

  /* scan character */
  for (ix = 0; ix < nr; ix++)
    { if (is_hex_digit (*ptr, &dig))
	{ value = value * 16 + dig;
	  ptr++;
	}
      else
	{ dat_error ("Hex digit expected in unicode sequence");
	  break;
	};
    };
  *sptr = ptr;

  /* Encode character */
  if (value & 0x1f0000)    { hdr = 0xF0; nr_bytes = 4; }
  else if (value & 0xF800) { hdr = 0xE0; nr_bytes = 3; }
  else if (value & 0x7800) { hdr = 0xC0; nr_bytes = 2; }
  else { hdr = 0; nr_bytes = 1; };

  do
    { nr_bytes--;
      *tptr++ = (char) (hdr | (value >> (nr_bytes * 6)));
      value &= ((1 << (nr_bytes * 6)) - 1);
      hdr = 0x80;
    }
  while (nr_bytes > 0);
  *dptr = tptr;
}

static char lexeme_buf[MAX_DAT_LINE_LEN + 1];
static int is_word_form (char **ret_ptr, int *ret_marker)
{ char string_buf[MAX_DAT_LINE_LEN + 1];
  char *sptr = string_buf;
  char *dptr = lexeme_buf;
  int marker = 0;
  int len;
  char ch;
  if (!is_string (string_buf)) return (0);
  len = (int) strlen (string_buf);
  
  /* Strip leading and trailing layout */
  while (ch = *sptr, ((ch == ' ') || (ch == '\t'))) { sptr++; len--; };
  while (ch = sptr[len - 1], ((ch == ' ') || (ch == '\t'))) sptr[--len] = '\0';

  /* Copy lexeme from source to lexeme_buf while expanding escapes */
  while ((ch = *sptr++))
    switch (ch)
      { case ' ':
	case '\t':
	  { /* Eat all other white space */
	    while ((ch = *sptr), (ch == ' ') || (ch == '\t')) sptr++;
	    *dptr++ = ' ';
	    marker |= LexemeMultiWordBit;
	  }; break;
	case '\\':
	  { ch = *sptr++;
	    switch (ch)
	      { case 'n': *dptr++ = '\n'; break;
		case 't': *dptr++ = '\t'; break;
		case 'r': *dptr++ = '\r'; break;
		case '\\': *dptr++ = '\\'; break;
		case '"': *dptr++ = '"'; break;
		case '!':
		  if (dptr != lexeme_buf)
		    { dat_warning ("Literal escape '\\!' is only recognized at start of lexeme");
		      *dptr++ = ch;
		    }
		  else if (marker & LexemeLiteralBit) /* "\!" is leading the lexeme */
		    dat_warning ("Multiple literal marking");
		  else marker |= LexemeLiteralBit;
		  break;
		case '-':
		  if ((dptr == lexeme_buf) && !(marker & LexemeSuffixBit))
		    /* "\-" is leading the lexeme, hence a suffix marker */
		    marker |= LexemeSuffixBit;
		  else if (!(*sptr))
		    /* "\-" is trailing the lexeme, hence a prefix marker */
		    marker |= LexemePrefixBit;
		  else
		    /* "\-" is in the middle of a lexeme and denotes the soft hyphen */
		    *dptr++ = SoftHyphenChar;
		  break;
		case 'u': add_utf8_character (&sptr, &dptr, 4); break;
		case 'U': add_utf8_character (&sptr, &dptr, 8); break;
		default:
		  dat_warning ("Unknown escape sequence '\\%c'", ch);
		  *dptr++ = ch;
	      };
	  }; break;
	default:
	  *dptr++ = ch;
      };
  *dptr = '\0';

  /* Check for empty lexeme */
  if (!strlen (lexeme_buf))
    dat_error ("Lexeme only consists of white space and literal/hyphen marks");

  /* Return with success */
  *ret_ptr = lexeme_buf;
  *ret_marker = marker;
  return (1);
}

/*
   Recognize a string is an actual parameter
   Note: we can always store it immediately in the parameter admin
*/
static int is_string_parameter (int *index)
{ char string_buf[MAX_DAT_LINE_LEN + 1];
  char cstring_buf[MAX_DAT_LINE_LEN + 1];
  char *sptr = string_buf;
  char *dptr = cstring_buf; 
  char ch;
  if (!is_string (string_buf)) return (0);

  /* Copy string from source to 2nd buffer while expanding escapes */
  while ((ch = *sptr++))
    switch (ch)
      { case '\\':
	  { ch = *sptr++;
	    switch (ch)
	      { case 'n': *dptr++ = '\n'; break;
		case 't': *dptr++ = '\t'; break;
		case 'r': *dptr++ = '\r'; break;

		/* To add: handling \u */
		/* To add: handling \x */
		default:
		  *dptr++ = ch;
	      };
	  }; break;
        default:
	  *dptr++ = ch;
      };
  *dptr = '\0';
  *index = register_new_text_affix (cstring_buf);
  return (1);
}

/*
   Recognition of parts of speech
   For the reading of affix names, we accept letters, digits and ornaments
*/
static int is_affix_name (int *index)
{ char buf[MAX_DAT_LINE_LEN+1];
  char *dptr = buf;
  while (!is_eoln () && ahead_name_char ())
    *dptr++ = *line_ptr++;
  *dptr = '\0';
  if (dptr == buf) return (0);
  may_skip_white_space ();
  *index = lookup_set_affix (buf);
  if (*index < 0)
    { dat_error ("Unknown set affix '%s'", buf);
      *index = FormalERROR;
    };
  return (1);
}

static void should_be_affix_name (int *index)
{ if (is_affix_name (index)) return;
  dat_error ("affix name expected");
  skip_eoln ();
  *index = FormalERROR;
}

static char nonterminal_buf[MAX_DAT_LINE_LEN + 1];
static int is_nonterminal_name ()
{ char *dptr = nonterminal_buf;
  if (!ahead_letter ()) return (0);
  while (!is_eoln () && ahead_name_char ())
    { *dptr++ = *line_ptr++;
      if (ch_is_white_space (*line_ptr))
        { may_skip_white_space ();

          /* Check for a new idpart */
	  if (is_eoln () || !ahead_letter ())
	    break;
          *dptr++ = ' ';
        };
    };
  *dptr = '\0';
  may_skip_white_space ();
  return (line_ptr != nonterminal_buf);
}

static void should_be_nonterminal_name ()
{ if (is_nonterminal_name ()) return;
  nonterminal_buf[0] = '\0';
  dat_error ("Nonterminal name expected");
  skip_eoln ();
}

/*
   Parameter recognization
   We store the formal typing, the affix indexes, and possible unions into
   three static separate arrays to avoid reallocating over and over again.
   The code for the punions array is quite dirty....
*/
#define DEFAULT_SIZE 256
#define DEFAULT_MAX_MARKERS 8
static int_array ptypes;
static int_array pidxs;
static ptr_array punions;
static void init_dat_parser ()
{ int ix;
  all_lexeme_markers = init_int_array (DEFAULT_MAX_MARKERS);
  all_vocabularies = init_ptr_array (DEFAULT_MAX_MARKERS);
  ptypes = init_int_array (DEFAULT_SIZE);
  pidxs = init_int_array (DEFAULT_SIZE);
  punions = init_ptr_array (DEFAULT_SIZE);

  /* Explicitly initialize the unions array */
  for (ix = 0; ix < DEFAULT_SIZE; ix++)
     punions -> array[ix] = (void *)NULL;
}

static void clear_parameter_admin ()
{ ptypes -> size = 0;
  pidxs -> size = 0;
  punions -> size = 0;
}

static int_array new_union_array (int *index)
{ int_array new_ints;
  if (punions -> size == punions -> room)
    { int ix;
      int new_room = punions -> room << 1;
      room_ptr_array (punions, new_room);
      for (ix = punions -> size; ix < new_room; ix++)
	punions -> array[ix] = (void *) NULL;
    };
  *index = punions -> size;
  if (punions -> array[punions -> size] == NULL)
    { new_ints = init_int_array (DEFAULT_SIZE);
      app_ptr_array (punions, (void *) new_ints);
      return (new_ints);
    };
  new_ints = (int_array) punions -> array[punions -> size];
  punions -> size++;
  new_ints -> size = 0;
  return (new_ints);
}

static int is_parameter ()
{ int_array new_union;
  int index, union_index;
  if (is_number_parameter (&index))
    { app_int_array (ptypes, FormalINT);
      app_int_array (pidxs, index);
      return (1);
    }
  else if (is_string_parameter (&index))
    { app_int_array (ptypes, FormalTEXT);
      app_int_array (pidxs, index);
      return (1);
    }
  if (!is_affix_name (&index)) return (0);
  if (!is_char ('|'))
    { app_int_array (ptypes, FormalSET);
      app_int_array (pidxs, index);
      return (1);
    };
  new_union = new_union_array (&union_index);
  app_int_array (ptypes, FormalUNION);
  app_int_array (pidxs, union_index);
  app_int_array (new_union, index);
  should_be_affix_name (&index);
  app_sorted_int_array (new_union, index);
  while (is_char ('|'))
    { should_be_affix_name (&index);
      app_sorted_int_array (new_union, index);
    };
  return (1);
}
 
static void should_be_parameter ()
{ if (is_parameter ()) return;
  dat_error ("parameter expected");
  skip_eoln ();
}

static void may_be_parameter_pack ()
{ if (is_char ('('))
    { should_be_parameter ();
      while (is_char (','))
	should_be_parameter ();
      should_be_char (')');
    };
  /* Register parameter pack */
}

static void type_check_union_and_unify (int ix, int formal)
{ int uni_nr = pidxs -> array[ix];
  int_array uni = (int_array) punions -> array[uni_nr]; 
  int part_error = 0;
  int aff_nr;
  int iy;

  /* Check if we have only one affix to unify */
  if (uni -> size == 1)
    { aff_nr = uni -> array[0];
      if (!affix_belongs_to_domain (aff_nr, formal))
	dat_error ("Type mismatch for parameter %d: union part %s against formal type %s",
                    ix + 1, affix_name_from_index (aff_nr), affix_name_from_index (formal));
      if (verbose)
	dat_warning ("Union affix resolved into single affix %s", affix_name_from_index (aff_nr));
      pidxs -> array[ix] = aff_nr;
      ptypes -> array[ix] = FormalSET;
      return;
    };

  /* Construct the union of the affixes */
  for (iy = 0; iy < uni -> size; iy++)
    { aff_nr = uni -> array[iy];
      if (!affix_belongs_to_domain (aff_nr, formal))
	{ dat_error ("Type mismatch for parameter %d: union part %s against formal type %s",
		     ix + 1, affix_name_from_index (aff_nr), affix_name_from_index (formal));
	  part_error = 1;
	};
    };

  /* Someone in error, resolve */
  if (part_error)
    { pidxs -> array[ix] = FormalERROR;
      ptypes -> array[ix] = FormalERROR;
    };

  /* Turn it into a single affix */
  aff_nr = register_new_union_affix (uni, formal);
  pidxs -> array[ix] = aff_nr;
  ptypes -> array[ix] = FormalSET;
}

static void type_check_parameter (int ix, int formal)
{ int ptype = ptypes -> array[ix];
  switch (ptype)
    { case FormalINT:
      case FormalTEXT:
	if (formal != ptypes -> array[ix])
	  dat_error ("Type mismatch for parameter %d: %s against formal type %s",
		     ix + 1, affix_name_from_index (ptype), affix_name_from_index (formal));
        break;
      case FormalSET:
	{ int aff_nr = pidxs -> array[ix];
	  if (formal < 0)
	    dat_error ("Type mismatch for parameter %d: %s against formal type %s",
		       ix + 1, affix_name_from_index (ptype), affix_name_from_index (formal));
	  else if (!affix_belongs_to_domain (aff_nr, formal))
	    dat_error ("Type mismatch for parameter %d: %s against formal type %s",
		       ix + 1, affix_name_from_index (aff_nr), affix_name_from_index (formal));
	}; break;
      case FormalUNION:
	{ if (formal < 0)
	    { dat_error ("Type mismatch for parameter %d: union type against formal type %s",
		         ix + 1, affix_name_from_index (formal));
	      break;
	    };
	  type_check_union_and_unify (ix, formal);
        };
      default: break;
    };
}

static void type_check_parameter_pack (int nont_index)
{ int_array formals = formals_from_nonterminal (nont_index);
  int ix;
  for (ix = 0; ix < formals -> size; ix++)
    type_check_parameter (ix, formals -> array[ix]);
}

static int *enter_into_lexicon (char *lexeme_ptr, int lexeme_marker)
{ Vocabulary new_voc = NULL;
  int *new_info_ptr;
  int ix;
  for (ix = 0; ix < all_lexeme_markers -> size; ix++)
    if (all_lexeme_markers -> array[ix] == lexeme_marker)
      { Vocabulary old_voc = (Vocabulary) all_vocabularies -> array[ix];
	return (lxcn_enter_into_vocabulary (&old_voc, lexeme_ptr));
      };

  /* We have encountered a new marker */
  new_info_ptr = lxcn_enter_into_vocabulary (&new_voc, lexeme_ptr);
  app_int_array (all_lexeme_markers, lexeme_marker);
  app_ptr_array (all_vocabularies, (void *) new_voc);
  return (new_info_ptr);
}

static int is_rule ()
{ char *lexeme_ptr;
  int lexeme_marker;
  int freq;
  if (is_word_form (&lexeme_ptr, &lexeme_marker))
    { /* Read remainder of rule */
      int *info_ptr;
      int nont_index, call_id;
      may_skip_white_space ();
      should_be_nonterminal_name ();
      may_be_parameter_pack ();
      may_be_signed_number (&freq);

      /* Identification and typecheck */
      nont_index = lookup_nonterminal (nonterminal_buf, ptypes -> size);
      if (nont_index < 0)
	{ dat_error ("No such nonterminal '%s/%d'", nonterminal_buf, ptypes -> size);
	  clear_parameter_admin ();
	  return (1);
	};

      /* Type check and register the entry */
      type_check_parameter_pack (nont_index);
      call_id = register_new_call (nont_index, pidxs);
      /* info_ptr = lxcn_enter_into_vocabulary (&the_vocabulary, lexeme_ptr); */
      info_ptr = enter_into_lexicon (lexeme_ptr, lexeme_marker);
      register_new_entry (info_ptr, call_id, freq);
      clear_parameter_admin ();
      return (1);
    };
  return (0);
}

/*
   Drive the lexicon parsing
*/
static void parse_dat_file (char *lname)
{ abs_message ("  reading lexicon %s...", lname);
  try_open_lexicon_file (lname);
  while (!is_eof ())
    { /* Body should eat line */
      may_skip_white_space ();
      if (is_eoln ()) read_line ();
      else if (is_comment ()) ;
      else if (is_rule ())
        { /* May still be followed by a comment */
	  if (is_comment ()) ;
          else should_be_eoln ();
	}
      else
	{ dat_error ("incomprehensible syntax");
	  read_line ();
        };
    }
  close_lexicon_file ();
}

static void report_lexicon_stats ()
{ abs_message ("    collected %d INT affixes, %d TEXT affixes, %d set affixes",
	       nr_of_int_affixes (), nr_of_text_affixes (), nr_of_set_affixes ());
  abs_message ("    collected %d different calls of lexicon nonterminals", nr_of_calls ());
  abs_message ("    collected %d lexicon entries, %d feature entries",
	       nr_of_entry_lists (), nr_of_entries ());
}

void parse_dat_files ()
{ int ix;
  init_dat_parser ();
  for (ix = 0; ix < lexicon_names -> size; ix++)
    parse_dat_file (lexicon_names -> array[ix]);
  if (verbose) report_lexicon_stats ();
}
